# coding=utf-8
# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A library of federated primitives defined as JAX primitives."""
from collections.abc import Callable
from typing import Protocol, Union

import jax
from jax import core
from jax import numpy as jnp
from jax.interpreters import ad
from jax.interpreters import batching
from jax.interpreters import mlir

from . import impls


class BroadcastType(Protocol):

  def __call__(
      self,
      x,
  ):
    Ellipsis


AggType = Callable[[jnp.ndarray], jnp.ndarray]


def _define_broadcast_prim(
    broadcast_name,
):
  """Defines and returns broadcast ptimitive and associated binding."""
  broadcast_p = core.Primitive(broadcast_name)  # Create the primitive

  def broadcast_prim_fn(x):
    return broadcast_p.bind(x)

  return (broadcast_p, broadcast_prim_fn)


def _register_broadcast_impls(
    broadcast_p,
    broadcast_prim_fn,
    broadcast_array_eval,
    sum_prim_fn,
    n_elements,
):
  """Registers implementations for the broadcast primitive.

  Definitions for Jacobian-vector products and vector-jacobian products are
  derived from forward and reverse differentiation rules presented in
  Federated Automatic Differentiation, https://arxiv.org/pdf/2301.07806v1.pdf.

  Args:
    broadcast_p: `core.Primitive` representing the broadcast operation.
    broadcast_prim_fn: A callable which binds its arguments to `broadcast_p`, as
      in
      https://jax.readthedocs.io/en/latest/notebooks/How_JAX_primitives_work.html#defining-new-jax-primitives
    broadcast_array_eval: A callable accepting and returning Jax arrays, which
      defines the array-valued implementation of this broadcast primitive.
    sum_prim_fn: A callable which binds its arguments to the summation primitive
      from the placement inserted by this broadcast. Similar to
      `broadcast_prim_fn`.
    n_elements: The number of elements present at the placement which this
      broadcast targets.
  """

  def broadcast_abstract_eval(xs):
    return core.ShapedArray((n_elements,) + xs.shape, xs.dtype)

  # Abstract eval rule.
  broadcast_p.def_abstract_eval(broadcast_abstract_eval)
  # Concrete eval rule.
  broadcast_p.def_impl(broadcast_array_eval)
  # Lowering rule to MLIR.
  mlir.register_lowering(
      broadcast_p, mlir.lower_fun(broadcast_array_eval, multiple_results=False)
  )

  def broadcast_jvp(primals_in, tangents_in):
    primals_out = broadcast_prim_fn(*primals_in)
    tangents_out = broadcast_prim_fn(*tangents_in)
    return primals_out, tangents_out

  # Registering JVP should allow forward AD.
  ad.primitive_jvps[broadcast_p] = broadcast_jvp

  def broadcast_vjp(cotangents_out, primals_in):
    if isinstance(cotangents_out, jax.interpreters.ad.Zero):
      # We are differerentiating back through a broadcast; the incoming value,
      # therefore, has the right shape and dtype for the Zero we generate.
      return (jax.interpreters.ad.Zero(primals_in.aval),)
    # This implementation *must* use the sum_prim_fn, rather than the array
    # implementation of summation, to result in a federated_sum in the Jaxpr.
    return (sum_prim_fn(cotangents_out),)

  ad.primitive_transposes[broadcast_p] = broadcast_vjp

  def _batch_broadcast(xs, batched_shape):
    # We inserted clients dimension in front, so batch dim went down one.
    return broadcast_prim_fn(*xs), batched_shape[0] + 1

  # Make sure this can also be batched / mapped. This happens when dispatching
  # forward AD, I think.
  batching.primitive_batchers[broadcast_p] = _batch_broadcast


def _define_single_arg_agg_prim(agg_name):
  """Defines and returns an aggregation primitive taking a single argument."""
  agg_p = core.Primitive(agg_name)  # Create the primitive

  def agg_prim_fn(x):
    return agg_p.bind(x)

  return agg_p, agg_prim_fn


def _register_single_arg_agg_impls(
    agg_p,
    agg_prim_fn,
    agg_array_eval,
    vjp_impl,
):
  """Registers implementations for aggregation primitive taking a single arg.

  Definitions for Jacobian-vector products and vector-jacobian products are
  derived from forward and reverse differentiation rules presented in
  Federated Automatic Differentiation, https://arxiv.org/pdf/2301.07806v1.pdf.

  Args:
    agg_p: `core.Primitive` representing the primitive aggregation operation.
    agg_prim_fn: A callable which binds its arguments to `agg_p`, as in
      https://jax.readthedocs.io/en/latest/notebooks/How_JAX_primitives_work.html#defining-new-jax-primitives
    agg_array_eval: A callable accepting and returning Jax arrays, which defines
      the array-valued implementation of this aggregation primitive.
    vjp_impl: An implementation of reverse-mode differentiation through this
      aggregation primitive.
  """

  def agg_abstract_eval(xs):
    return jax.tree_util.tree_map(
        # We slice away the first dimension in doing the reduction; its gone!
        lambda x: core.ShapedArray(x.shape[1:], x.dtype),
        xs,
    )

  # Abstract eval rule
  agg_p.def_abstract_eval(agg_abstract_eval)
  # Concrete eval rule
  agg_p.def_impl(agg_array_eval)
  # Lowering rule to MLIR.
  mlir.register_lowering(
      agg_p,
      mlir.lower_fun(agg_array_eval, multiple_results=False),
  )

  def agg_jvp(primals_in, tangents_in):
    return agg_prim_fn(*primals_in), agg_prim_fn(*tangents_in)

  # Registering JVP should allow forward AD.
  ad.primitive_jvps[agg_p] = agg_jvp

  def agg_vjp(cotangents_out, primals_in):
    if isinstance(cotangents_out, jax.interpreters.ad.Zero):
      # We are differerentiating back through an aggregation; the incoming
      # value, therefore, has the right shape and dtype for the Zero we
      # generate. This is always correct if jax's symbolic Zero is a static
      # concept, depending on data flow in the program (rather than e.g. runtime
      # values).
      return (jax.interpreters.ad.Zero(primals_in.aval),)
    return (vjp_impl(cotangents_out),)

  ad.primitive_transposes[agg_p] = agg_vjp

  def _batch_agg(xs, batched_shape):
    # Certain jax libs can silently insert the 'batching' dim 'all the way at
    # the front'; we are about to destroy the front axis by agging, so move
    # that puppy to the back. Tell the rest of JAX what happened here.
    xs = batching.moveaxis(*xs, *batched_shape, -1)
    return agg_prim_fn(xs), len(xs.shape) - 2

  # Make sure this can also be batched / mapped. This happens when dispatching
  # forward AD, I think.
  batching.primitive_batchers[agg_p] = _batch_agg


def _define_and_register_prims_for_placement(
    primitive_dict,
    primdef_dict,
    impl_defs,
    placement_str,
    n_elements,
):
  """Registers primitives for a given placement and cardinality.

  Args:
    primitive_dict: Dictionary in which to place the defined primitive
      _functions_.
    primdef_dict: Dictionary in which to place the defined primitive _objects_.
    impl_defs: Instance of `impls.PlacedComputations` which provides the
      Jax-array accepting and returning implementations for the primitives we
      define here.
    placement_str: String representing the name of the placement to which we may
      broadcast, and from which we may aggregate.
    n_elements: The cardinality of this placement. That is, how many
      participants this placement supports.
  """
  broadcast_name = f'broadcast_{placement_str}'
  sum_name = f'sum_from_{placement_str}'
  mean_name = f'mean_from_{placement_str}'
  # Primitive and prim_fn definitions must come first, since they may need to
  # generate other primitives in the body of their autodiff implementations.
  broadcast_p, broadcast_prim_fn = _define_broadcast_prim(broadcast_name)
  sum_p, sum_prim_fn = _define_single_arg_agg_prim(sum_name)
  mean_p, mean_prim_fn = _define_single_arg_agg_prim(mean_name)

  primitive_dict[broadcast_name] = broadcast_prim_fn
  primitive_dict[sum_name] = sum_prim_fn
  primitive_dict[mean_name] = mean_prim_fn

  primdef_dict[broadcast_name] = broadcast_p
  primdef_dict[sum_name] = sum_p
  primdef_dict[mean_name] = mean_p

  def broadcast_array_eval(x):
    return impl_defs.broadcast_to_placement(x, placement_str)

  _register_broadcast_impls(
      broadcast_p,
      broadcast_prim_fn,
      broadcast_array_eval,
      sum_prim_fn,
      n_elements,
  )

  _register_single_arg_agg_impls(
      sum_p,
      sum_prim_fn,
      impl_defs.sum_from_placement,
      broadcast_prim_fn,
  )
  _register_single_arg_agg_impls(
      mean_p,
      mean_prim_fn,
      impl_defs.mean_from_placement,
      # TODO(b/277797059): Similar comment here to the sum.
      lambda x: jnp.divide(broadcast_prim_fn(x), n_elements),
  )


def register_primitives(
    placements,
):
  """Registers broadcast, sum and mean primitives for all placements present."""
  impl_defs = impls.PlacedComputations(placements)
  primitive_dict = {}
  primdef_dict = {}
  for placement_str, n_elements in placements.items():
    _define_and_register_prims_for_placement(
        primitive_dict, primdef_dict, impl_defs, placement_str, n_elements
    )

  return primitive_dict, primdef_dict



# coding=utf-8
# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The FAX package."""

import functools as _functools
import sys as _sys

from . import api as _api


# Import the public API.
federated_broadcast = _api.federated_broadcast
federated_map_clients = _api.federated_map_clients
federated_map_server = _api.federated_map_server
federated_mean = _api.federated_mean
federated_sum = _api.federated_sum
federated_weighted_mean = _api.federated_weighted_mean


@_functools.wraps(_api.fax_program)
def fax_program(*, placements):
  # We wrap here and send in this module as the one to be modified, as it
  # will be the one that users interact with and requires the API changes.
  return _api.fax_program(
      placements=placements, self_module=_sys.modules[__name__]
  )



# coding=utf-8
# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An API for embedding a federated programming model into JAX.

This is done by patching a global module during decorated-function invocation to
give the appearance a single, uniform API. This could certainly be changed
by altering the method of dispatch--e.g., by letting the data we carry through
function invocations drive which concrete function is called.
"""

from collections.abc import Callable
import functools
from typing import Any

from absl import logging
import jax

from . import impls
from . import primitives

# We define type aliases to make specifying the API below easier.
_NestedJTensor = Any
_NestedServerTensor = Any
_NestedClientsTensor = Any
_NestedPSpec = Any


# We don't want to make the degenerate case of directly calling these symbols
# outside of a fax decorator part of the API.
# pylint: disable=g-doc-exception
_FEDERATED_PRIMITIVES = {}


class OperatorUndefinedError(Exception):

  def __init__(self, name):
    self.message = (
        f'The FAX operator {name} is only defined in the context of a fax '
        'program decorator. Attempted to call without using this decorator.'
    )
    super().__init__(self.message)


# The following functions which 'do nothing but raise' can be considered
# something like a specification of FAX's API.


# The 'transport' symbols (broadcast, sum, mean, etc) can be considered the
# core of fax: everything else is 'just helpers'. FAX's main job is to capture
# the transport of arrays at a high level, and provide implementations which
# manage the triplet of 'federated semantics / logical representation in
# program / physical layout'.
def federated_broadcast(
    x,
):
  """Broadcasts its input to the clients.

  The broadcast tensor will be partitioned along the 'clients' dimension of
  the
  mesh. Other dimensions will be unconstrained, and are expected to be laid
  out
  well by the XLA compiler (e.g., preserving the sharding of input tensors).

  Args:
    x: A structure of arrays to be broadcast.

  Returns:
    A structure of arrays broadcast to the clients dimension (i.e., each array
    has a clients-dimension inserted on the left, of size configured in the
    FAX
    decorator), laid out as described above.
  """
  raise OperatorUndefinedError('federated_broadcast')


def federated_mean(x):
  """Computes a federated (unweighted) mean across clients.

  At runtime in the datacenter, this function simply runs as a `jnp.mean`
  across
  the client dimension.

  Args:
    x: The (potentially nested) clients-placed tensor to be federated_mean'ed.

  Returns:
    The result of an unweighted mean across the clients dimension.
  """
  raise OperatorUndefinedError('federated_mean')


def federated_sum(x):
  """Computes a federated sum across clients.

  At runtime in the datacenter, this function simply runs as a `jnp.sum`
  across
  the client dimension.

  Args:
    x: The (potentially nested) clients-placed tensor to be federated_sum'ed.

  Returns:
    The result of sum across the clients dimension.
  """
  raise OperatorUndefinedError('federated_sum')


def federated_weighted_mean(
    x, w
):
  """Computes a federated weighted mean across clients.

  At runtime in the datacenter, this function delegates internally to two
  `federated_sum`s, to compute numerator and denominator.

  Args:
    x: The (potentially nested) clients-placed tensor to be weighted-mean'ed.
    w: The weight to use for this average.

  Returns:
    The result of the weighted mean across the clients dimension.
  """
  raise OperatorUndefinedError('federated_weighted_mean')


# The mapping functions here are provided simply for convenience. Users can
# write *any* processing across their federated arrays; but we provide our
# mapping functions to ease the burden of writing sharded + distributed jax.
# In particular, `federated_map_clients` is intended as a location to
# centralize and codify knowledge on writing code which enables weak scaling
# with respect to clients, a more subtle question than it might seem.


def federated_map_clients(
    fn, arg
):
  """Maps `fn` across the clients dimension of `arg`.

  This function will direct the GSPMD compiler that the mapped axis
  corresponds
  to the 'clients' dimension of the mesh installed at tracing time, if such an
  axis exists.

  Args:
    fn: A callable accepting slices along the clients dimension of `arg`.
    arg: A clients-placed structure of tensors to be mapped. If `arg` is a
      tuple, this tuple will be unpacked upon invoking `fn`.

  Returns:
    The result of mapping the function `fn` as described above.
  """
  raise OperatorUndefinedError('federated_map_clients')


def federated_map_server(
    fn, arg
):
  """Calls `fn` on `arg`.

  Since fax represents server-placed arrays as *identical* to unplaced
  arrays, this function simply calls `fn` on `arg`, and users are free to
  replace this function with its body.

  Args:
    fn: The function to apply.
    arg: The (singleton) argument to pass to `fn`. If `arg` is a tuple, this
      tuple will be unpacked while invoking `fn`.

  Returns: The result of calling `fn` on `arg`.
  """
  raise OperatorUndefinedError('federated_map_server')


# pylint: disable=g-long-lambda
def _replace_api(api, placed_computations, prim_computations):
  # importlib.util.module_from_spec()
  """A binding of the implementations in FAX to represent FL."""
  api.federated_map_clients = (
      lambda fn, arg: placed_computations.map_to_placement(fn, arg, 'clients')
  )
  api.federated_map_server = (
      lambda fn, arg: placed_computations.map_to_placement(fn, arg, 'server')
  )

  api.federated_mean = lambda x: jax.tree_util.tree_map(
      prim_computations['mean_from_clients'], x
  )
  api.federated_sum = lambda x: jax.tree_util.tree_map(
      prim_computations['sum_from_clients'], x
  )

  def _weighted_mean(x, w):
    mult_at_clients = api.federated_map_clients(
        lambda arg1, arg2: jax.tree_util.tree_map(
            lambda x, y: x * y, arg1, arg2
        ),
        (x, w),
    )
    sum_mult = api.federated_sum(mult_at_clients)
    denom_sum = api.federated_sum(w)
    return jax.tree_util.tree_map(lambda x, y: x / y, sum_mult, denom_sum)

  api.federated_weighted_mean = _weighted_mean

  def _broadcast_fn(x, *, input_pspecs=None):
    # TODO: b/308448854 - Remove this parameter and update callsites.
    del input_pspecs  # Unused

    return jax.tree_util.tree_map(prim_computations['broadcast_clients'], x)

  api.federated_broadcast = _broadcast_fn
  return api


# pylint: enable=g-long-lambda


def fax_program(*, placements, self_module):
  """Patches symbols into current module and call `jax.jit` on the result.

  This decorator enables calling:

  * federated_map_at_clients
  * federated_map_at_server
  * federated_mean
  * federated_sum
  * federated_weighted_mean
  * federated_broadcast

  The functions returned by this decorator has been jit compiled, and are
  compatible with `jax.jacfwd` and `jax.jacrev`.


  Args:
    placements: Dictionary defining placements. Must contain specification of
      number of clients. Other placements may be ignored inside this decorator.
    self_module: The Python module to patch the API when performing FAX tracing.

  Returns:
    A decorated function enabling the calling of the FAX API. Interoperable
    with other JAX code.
  """
  if 'clients' not in placements:
    raise ValueError(
        'Need a specification for cardinalities of clients placement.'
    )

  placed_computations = impls.PlacedComputations(
      placements_to_n_elements=placements,
  )
  prim_computations, primdefs = primitives.register_primitives(
      placements=placements
  )
  global _FEDERATED_PRIMITIVES
  if not _FEDERATED_PRIMITIVES:
    # First time these primitives are defined; attach them to the global.
    _FEDERATED_PRIMITIVES = primdefs

  def fn_decorator(fn):
    @functools.wraps(fn)
    def jax_callable(*args, **kwargs):
      old_api_symbols = {
          attribute_name: getattr(self_module, attribute_name)
          for attribute_name in dir(self_module)
      }
      # We need to patch down here so that the values that get traced at
      # 'runtime' are the expected ones.
      try:
        _replace_api(self_module, placed_computations, prim_computations)
        logging.info('lib patched!')
        return fn(*args, **kwargs)
      finally:
        for symbol_name, symbol in old_api_symbols.items():
          setattr(self_module, symbol_name, symbol)
        logging.info('lib patch reverted!')

    return jax_callable

  return fn_decorator


# pylint: enable=g-doc-exception



# coding=utf-8
# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import chex
import jax
from jax import numpy as jnp
import tensorflow as tf

from . import impls


class ImplsTest(tf.test.TestCase):

  def setUp(self):
    super().setUp()
    self._placements = {'clients': 100}
    self._sequence_length = 10
    self._comp_factory = impls.PlacedComputations(
        placements_to_n_elements=self._placements,
    )

  def test_map_to_server_placement_applies_fn(self):
    def add_constant(x):
      return x + jnp.array([1.0, 2.0])

    server_val = jnp.array([0.0, 1.0])
    actual_output = self._comp_factory.map_to_placement(
        add_constant, server_val, 'server'
    )
    expected_output = add_constant(server_val)
    chex.assert_trees_all_equal(actual_output, expected_output)

  def test_broadcast_on_float(self):
    actual_output = self._comp_factory.broadcast_to_placement(0.0, 'clients')
    expected_output = jnp.zeros(shape=[100])
    chex.assert_trees_all_equal(actual_output, expected_output)

  def test_runs_temp_sens_example(self):
    def _one_if_over(x, y):
      return jax.lax.cond(x > y, lambda: 1.0, lambda: 0.0)

    def temp_sens_example(m, t):
      t_at_c = self._comp_factory.broadcast_to_placement(t, 'clients')
      total_over = self._comp_factory.map_to_placement(
          _one_if_over, (m, t_at_c), 'clients'
      )
      return self._comp_factory.mean_from_placement(total_over)

    key = jax.random.PRNGKey(2)
    random_measurements = jax.random.uniform(
        key, shape=[self._placements['clients']]
    )

    self.assertEqual(
        temp_sens_example(random_measurements, jnp.array(0.5)), 0.53
    )

  def test_runs_fake_training(self):
    def _reduce_sequence(sequence, model):

      for _ in range(sequence.shape[0]):
        model += 1
      return model

    def fake_training(model, data):
      model_at_clients = self._comp_factory.broadcast_to_placement(
          model, 'clients'
      )
      trained_models = self._comp_factory.map_to_placement(
          _reduce_sequence, (data, model_at_clients), 'clients'
      )
      return self._comp_factory.mean_from_placement(trained_models)

    clients_data = jnp.ones(
        shape=[self._placements['clients'], self._sequence_length]
    )

    model = jnp.array(0.0)

    self.assertEqual(fake_training(model, clients_data), 10.0)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import jax
from jax import numpy as jnp
import numpy as np
import tensorflow as tf

from . import api


@functools.wraps(api.fax_program)
def fax_program(*, placements):
  return api.fax_program(placements=placements, self_module=api)


class ApiTest(tf.test.TestCase):

  def test_sharded_broadcast(self):

    @fax_program(placements={"clients": 100})
    def broadcast_val(val):
      return api.federated_broadcast(val)

    mesh = jax.sharding.Mesh(np.array(jax.devices()), ("some_axis",))
    arg_sharding = jax.sharding.NamedSharding(
        mesh, jax.sharding.PartitionSpec("some_axis")
    )
    with mesh:
      result = broadcast_val(
          jax.device_put(jnp.ones(shape=[8, 8]), arg_sharding)
      )

    self.assertAllClose(result, jnp.ones(shape=[100, 8, 8]))

    self.skipTest("Following assertion fails with an empty PartitionSpec.")
    # No clients dimension in the mesh, we don't lay out the clients along that
    # nonexistent dimension, but rather replicate them. Notice that we don't
    # need to specify the sharding to FAX; it should be inferred by GSPMD.
    expected_result_pspec = jax.sharding.PartitionSpec(None, "some_axis")
    self.assertEqual(
        result.sharding, jax.sharding.NamedSharding(mesh, expected_result_pspec)
    )

  def test_temp_sens_example(self):
    def one_if_over(threshold, value):
      return jax.lax.cond(value > threshold, lambda: 1.0, lambda: 0.0)

    @fax_program(placements={"clients": 100})
    def temp_sens_example(threshold, values):
      threshold_at_clients = api.federated_broadcast(threshold)
      values_over = api.federated_map_clients(
          one_if_over, (threshold_at_clients, values)
      )
      return api.federated_mean(values_over)

    key = jax.random.PRNGKey(2)
    random_measurements = jax.random.uniform(key, shape=[100])

    self.assertEqual(
        temp_sens_example(jnp.array(0.5), random_measurements), 0.53
    )

  def test_temp_sens_example_multiple_clients(self):
    def one_if_over(threshold, value):
      return jax.lax.cond(value > threshold, lambda: 1.0, lambda: 0.0)

    @fax_program(placements={"clients": 100})
    def temp_sens_example_100_clients(threshold, values):
      threshold_at_clients = api.federated_broadcast(threshold)
      values_over = api.federated_map_clients(
          one_if_over, (threshold_at_clients, values)
      )

      return api.federated_mean(values_over)

    @fax_program(placements={"clients": 10})
    def temp_sens_example_10_clients(threshold, values):
      threshold_at_clients = api.federated_broadcast(threshold)
      values_over = api.federated_map_clients(
          one_if_over, (threshold_at_clients, values)
      )
      return api.federated_mean(values_over)

    key = jax.random.PRNGKey(2)
    random_measurements_100 = jax.random.uniform(key, shape=[100])
    random_measurements_10 = jax.random.uniform(key, shape=[10])

    self.assertEqual(
        temp_sens_example_100_clients(jnp.array(0.5), random_measurements_100),
        0.53,
    )
    self.assertEqual(
        temp_sens_example_10_clients(jnp.array(0.5), random_measurements_10),
        0.4,
    )
    # We should be able to recover the original result flipping back to the
    # original function.
    self.assertEqual(
        temp_sens_example_100_clients(jnp.array(0.5), random_measurements_100),
        0.53,
    )

  def test_raises_outside_program_context(self):
    with self.assertRaises(api.OperatorUndefinedError):
      api.federated_broadcast(jnp.array(0.5))

    num_clients = 10

    @fax_program(placements={"clients": num_clients})
    def test(values):
      return api.federated_mean(values)

    # Should not raise, inside a fax context.
    test(jax.random.uniform(jax.random.PRNGKey(42), shape=[num_clients]))

    # Should raise again now that we've left the context.
    with self.assertRaises(api.OperatorUndefinedError):
      api.federated_broadcast(jnp.array(0.5))


if __name__ == "__main__":
  tf.test.main()



# coding=utf-8
# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the sharding behavior of the implementation of FAX primitives."""

import math
import unittest

from absl.testing import absltest
from absl.testing import parameterized
import jax
from jax import numpy as jnp
import numpy as np

from . import impls


# We program FAX to expect a mesh axis named _CLIENTS_AXIS. Consider defining +
# exporting a constant to help make mesh construction easier and guarantees
# cleaner. Other constants defined here are intended to help make the operations
# of the assertions in the tests below more transparent.
_CLIENTS_AXIS = 'clients'
_CLIENTS_AXIS_SIZE = 2
_NUM_CLIENTS = 100

_DATA_AXIS = 'data'
_DATA_AXIS_SIZE = 2
_DATA_SIZE = 10


# Inline a helper function for creating and manipulating test meshes from
# JAX's internals.
def create_global_mesh(mesh_shape, axis_names):
  size = math.prod(mesh_shape)
  if len(jax.devices()) < size:
    raise unittest.SkipTest(f'Test requires {size} global devices.')
  devices = sorted(jax.devices(), key=lambda d: d.id)
  mesh_devices = np.array(devices[:size]).reshape(mesh_shape)
  global_mesh = jax.sharding.Mesh(mesh_devices, axis_names)
  return global_mesh


class BroadcastShardingBehaviorTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._placements = {_CLIENTS_AXIS: _NUM_CLIENTS}
    self._comp_factory = impls.PlacedComputations(
        placements_to_n_elements=self._placements,
    )

  def test_broadcast_with_1x1_fully_replicates(self):
    arg = jnp.zeros(shape=[_DATA_SIZE])
    with create_global_mesh([1, 1], [_CLIENTS_AXIS, _DATA_AXIS]):
      result = self._comp_factory.broadcast_to_placement(arg, _CLIENTS_AXIS)
    self.assertEqual(result.shape, (_NUM_CLIENTS, _DATA_SIZE))
    sharding = result.sharding
    # There is only one chip we talk to, so this sharding 'looks' fully
    # replicated.
    self.assertTrue(sharding.is_fully_replicated)

  def test_broadcast_clients_shards_along_clients(self):
    arg = jnp.zeros(shape=[_DATA_SIZE])
    with create_global_mesh(
        [_CLIENTS_AXIS_SIZE, _DATA_AXIS_SIZE], [_CLIENTS_AXIS, _DATA_AXIS]
    ):
      result = self._comp_factory.broadcast_to_placement(arg, _CLIENTS_AXIS)
    self.assertEqual(result.shape, (_NUM_CLIENTS, _DATA_SIZE))
    sharding = result.sharding
    # If this sharding were fully replicated, we would be *replicating* the data
    # on each chip, rather than putting half of the clients' broadcasted arrays
    # on one set of client chips and half on the other.
    self.assertFalse(sharding.is_fully_replicated)
    # Each shard should host half the clients, but the arg's original dimension
    # should be replicated.
    self.assertEqual(
        sharding.shard_shape(result.shape),
        (_NUM_CLIENTS // _CLIENTS_AXIS_SIZE, _DATA_SIZE),
    )

  def test_broadcast_preserves_sharding_with_no_clients_mesh(self):
    arg = jnp.zeros(shape=[_DATA_SIZE])
    # Replicating a situation in which the caller's mesh has no clients axis; in
    # this case, we should preserve the sharding of any broadcast tensors, but
    # not shard along the (nonexistent) clients axis.
    no_mesh_comp_factory = impls.PlacedComputations(
        placements_to_n_elements=self._placements,
    )
    with create_global_mesh([_DATA_AXIS_SIZE], [_DATA_AXIS]) as mesh:
      arg_spec = jax.sharding.PartitionSpec('data')
      sharded_arg = jax.device_put(
          arg, device=jax.sharding.NamedSharding(mesh, arg_spec)
      )
      result = no_mesh_comp_factory.broadcast_to_placement(
          sharded_arg, _CLIENTS_AXIS
      )
    self.assertEqual(result.shape, (_NUM_CLIENTS, _DATA_SIZE))
    sharding = result.sharding
    self.assertIsInstance(sharding, jax.sharding.NamedSharding)
    # The data should be partitioned across chips.
    self.assertFalse(sharding.is_fully_replicated)
    # The resulting broadcast array should have the same sharding as its input
    # for the non-injected dimensions, and replication on the clients dimension.
    self.assertEqual(sharding.spec, jax.sharding.PartitionSpec(None, 'data'))
    # Here, the clients axis should be replicated on each set of chips;
    # however,the data making up the broadcasted array should be split along
    # the 'data' dimension; thus only the second dimension of the tensor should
    # be split.
    self.assertEqual(
        sharding.shard_shape(result.shape),
        (_NUM_CLIENTS, _DATA_SIZE // _DATA_AXIS_SIZE),
    )

  def test_broadcast_preserves_arg_sharding_with_clients_mesh(self):
    arg = jnp.zeros(shape=[_DATA_SIZE])
    with create_global_mesh(
        [_CLIENTS_AXIS_SIZE, _DATA_AXIS_SIZE], [_CLIENTS_AXIS, _DATA_AXIS]
    ) as mesh:
      arg_spec = jax.sharding.PartitionSpec('data')
      sharded_arg = jax.device_put(
          arg, device=jax.sharding.NamedSharding(mesh, arg_spec)
      )
      result = self._comp_factory.broadcast_to_placement(
          sharded_arg, _CLIENTS_AXIS
      )
    self.assertEqual(result.shape, (_NUM_CLIENTS, _DATA_SIZE))
    sharding = result.sharding
    self.assertIsInstance(sharding, jax.sharding.NamedSharding)
    # The data should be partitioned across chips.
    self.assertFalse(sharding.is_fully_replicated)
    # The resulting broadcast array should have the same sharding as its input
    # for the non-injected dimensions, and replication on the clients dimension.
    self.assertEqual(
        sharding.spec, jax.sharding.PartitionSpec(_CLIENTS_AXIS, _DATA_AXIS)
    )
    # Similarly to the above, the clients dimension should be 'split' across the
    # two sets of chips making up the clients axis of the mesh. However, in this
    # case the data making up the broadcasted array should *also* be split along
    # the 'data' dimension; thus each dimension of the global array's shape
    # should be cut in half, with sub-arrays of shape (_NUM_CLIENTS //
    # _CLIENTS_AXIS_SIZE,
    # _DATA_SIZE // _DATA_AXIS_SIZE) living on each of
    # the 4 chips.
    self.assertEqual(
        sharding.shard_shape(result.shape),
        (_NUM_CLIENTS // _CLIENTS_AXIS_SIZE, _DATA_SIZE // _DATA_AXIS_SIZE),
    )


@jax.jit
def add(x, y):
  # A computation that can be sharded out. Here we rely on the
  # 'computation-follows-data' behavior of jax.jit.
  return x + y


def _place_args_at_clients(*args, comp_factory):
  return tuple(
      comp_factory.broadcast_to_placement(x, _CLIENTS_AXIS) for x in args
  )


class MapFnShardingTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._placements = {_CLIENTS_AXIS: _NUM_CLIENTS}
    self._sequence_length = 10
    self._comp_factory = impls.PlacedComputations(
        placements_to_n_elements=self._placements,
    )

  def test_map_respects_clients_sharding(self):
    with create_global_mesh(
        [_CLIENTS_AXIS_SIZE, _DATA_AXIS_SIZE], [_CLIENTS_AXIS, _DATA_AXIS]
    ):
      arg1_at_c, arg2_at_c = _place_args_at_clients(
          jnp.zeros(shape=[_DATA_SIZE]),
          jnp.ones(shape=[_DATA_SIZE]),
          comp_factory=self._comp_factory,
      )
      result = self._comp_factory.map_to_placement(
          add, (arg1_at_c, arg2_at_c), _CLIENTS_AXIS
      )
    self.assertEqual(result.shape, (_NUM_CLIENTS, _DATA_SIZE))
    sharding = result.sharding
    # The data should be partitioned across chips.
    self.assertFalse(sharding.is_fully_replicated)
    # Each shard should host half the clients, but the arg's original dimension
    # should be replicated, since the arguments were replicated.
    self.assertEqual(
        sharding.shard_shape(result.shape),
        (_NUM_CLIENTS // _CLIENTS_AXIS_SIZE, _DATA_SIZE),
    )

  def test_map_respects_non_clients_sharding(self):
    arg_spec = jax.sharding.PartitionSpec('data')
    with create_global_mesh(
        [_CLIENTS_AXIS_SIZE, _DATA_AXIS_SIZE], [_CLIENTS_AXIS, _DATA_AXIS]
    ) as mesh:
      sharded_arg1 = jax.device_put(
          jnp.zeros(shape=[_DATA_SIZE]),
          device=jax.sharding.NamedSharding(mesh, arg_spec),
      )
      sharded_arg2 = jax.device_put(
          jnp.ones(shape=[_DATA_SIZE]),
          device=jax.sharding.NamedSharding(mesh, arg_spec),
      )
      arg1_at_c, arg2_at_c = _place_args_at_clients(
          sharded_arg1,
          sharded_arg2,
          comp_factory=self._comp_factory,
      )
      result = self._comp_factory.map_to_placement(
          add, (arg1_at_c, arg2_at_c), _CLIENTS_AXIS
      )
    self.assertEqual(result.shape, (_NUM_CLIENTS, _DATA_SIZE))
    # The data should be partitioned across chips.
    sharding = result.sharding
    self.assertIsInstance(sharding, jax.sharding.NamedSharding)
    self.assertFalse(sharding.is_fully_replicated)
    # The resulting array here should be fully sharded, just like the argument,
    # by computation-follows-data semantics.
    self.assertEqual(
        sharding.spec, jax.sharding.PartitionSpec(_CLIENTS_AXIS, 'data')
    )
    # Since the argument was fully split across the data and clients axes, the
    # result should be too: each of the 4 chips hosts a sub-array slice of data,
    # of shape (_NUM_CLIENTS // _CLIENTS_AXIS_SIZE, _DATA_SIZE //
    # _DATA_AXIS_SIZE), so that the entire (global) shape is (_NUM_CLIENTS,
    # _DATA_SIZE).
    self.assertEqual(
        sharding.shard_shape(result.shape),
        (_NUM_CLIENTS // _CLIENTS_AXIS_SIZE, _DATA_SIZE / _DATA_AXIS_SIZE),
    )


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sketching one potential interface."""
from typing import Any

import jax
import jax.experimental.maps
from jax.experimental.shard_alike import shard_alike
import jax.numpy as jnp
from jax.sharding import PartitionSpec as P


# The type aliases below encode a federated programming model, inpsired by
# the programming model of TensorFlow Federated (TFF). Notable differences
# are that placements are static, and can only appear on arrays. There are no
# nested placements. Placement dimension comes first, so there is no need to
# worry about a total order on placements or figuring out what representation a
# nested-placed thing should have.

UnplacedArray = jnp.ndarray
PlacedArray = jnp.ndarray

# Special alias for 'arrays with an extra sequence dimension in the first
# element'.
SequenceArray = jnp.ndarray
PyTreeUnplaced = Any
PyTreePlaced = Any
PyTreeSequence = Any


SERVER_PLACEMENT = 'server'


def call_jaxpr(fn, arg):
  # Handles multi-element arguments.
  if isinstance(arg, tuple):
    return fn(*arg)
  else:
    return fn(arg)


def read_sequence_batch(
    array_sequence, index
):
  sequence_at_idx = jax.tree_util.tree_map(
      lambda x: x[index, Ellipsis], array_sequence
  )
  return sequence_at_idx


def _global_mesh_defined():
  """Checks if global xmap/pjit mesh resource environment is defined."""
  maps_env = jax.experimental.maps.thread_resources.env
  return maps_env.physical_mesh.devices.shape != ()  # pylint: disable=g-explicit-bool-comparison


def _placement_axis_in_mesh(placement):
  """Checks if a clients axis is present in the global mesh."""
  if not _global_mesh_defined():
    return False
  mesh = jax.experimental.maps.thread_resources.env.physical_mesh
  return placement in mesh.axis_names


def _constrain_if_mesh(x, pspec):
  if not _global_mesh_defined():
    return x
  return jax.lax.with_sharding_constraint(x, pspec)


class PlacedComputations:
  """Concrete implementations of federated primitives in JAX."""

  def __init__(
      self,
      placements_to_n_elements,
  ):
    self._placements_to_n_elements = placements_to_n_elements

  def broadcast_to_placement(
      self,
      arg,
      placement,
  ):
    """Broadcasts (replicates) to the specified placement.

    That is, given an `arg` of shape `[a, ... b]`, and a `placement` with `n`
    elements, the result of this function should be an array of shape
    `[n, a, ... b]`, each of whose slices on the zeroth axis is identical to
    `arg`.

    This function shards the resulting replicated array along a mesh axis
    corresponding to 'placement' if such a mesh is defined at compilation time.
    Otherwise no sharding is applied and the result is fully replicated.

    This function must also direct the GSPMD compiler to shard the
    zeroth-axis slices of this replicated array in a similar manner to the
    argument.

    Args:
      arg: An array to be broadcast.
      placement: String representing the placement to which to broadcast `arg`.

    Returns:
      A logically replicated array along the zeroth axis, as described above.
    """
    arg = jnp.array(arg)
    n_elements = self._placements_to_n_elements[placement]

    # Note that this pspec will only result in a sharding constraint defined if
    # a mesh is installed at tracing time.
    if _placement_axis_in_mesh(placement):
      pspec = P(placement, *([P.UNCONSTRAINED] * len(arg.shape)))
    else:
      # Without a clients axis in the mesh, we simply explicitly tell the
      # compiler that there are no constraints on this tensor. This will leave
      # the choices in the hands of the compiler.
      pspec = P(*([P.UNCONSTRAINED] * (len(arg.shape) + 1)))

    def single_arg_broadcast(x):
      replicated_tensor = jnp.tile(x, reps=[n_elements] + [1] * len(x.shape))
      if not _global_mesh_defined():
        # No sharding expected, don't worry about it.
        return replicated_tensor
      else:

        def _shard_slice_like_arg(s):
          s_sharded, _ = shard_alike(s, x)
          return s_sharded

        original_dims_constrained = jax.vmap(_shard_slice_like_arg, in_axes=0)(
            replicated_tensor
        )
        fully_constrained = _constrain_if_mesh(original_dims_constrained, pspec)
        return fully_constrained

    return jax.jit(single_arg_broadcast)(arg)

  def normalized_broadcast_to_placement(
      self,
      arg,
      placement,
  ):
    # This broadcasts arg / placement_size. This is intended only for reverse-
    # mode differentiation of mean-based aggregations.
    n_elements = self._placements_to_n_elements[placement]
    unnormalized_broadcast = self.broadcast_to_placement(arg, placement)
    return jnp.divide(unnormalized_broadcast, n_elements)

  def mean_from_placement(self, arg):
    placement_idx = 0
    return jnp.mean(arg, axis=[placement_idx])

  def weighted_mean_from_placement(
      self, arg, weight
  ):
    placement_idx = 0
    return jnp.average(arg, axis=[placement_idx], weights=weight)

  def sum_from_placement(self, arg):
    placement_idx = 0
    return jnp.sum(arg, axis=[placement_idx])

  def map_to_placement(self, fn, arg, placement):
    """Maps a function to the specified placement.

    Suppose the user has a mesh with three axes, [placement, 'y', z']. Suppose
    we are asked to map a function f, of signature ([a], [b]) -> [c]. Suppose we
    have an array e, of shape [d, a, b], layed out along the mesh's three axes.

    Now, our mapping implementation is required to be able to: map f across the
    axis of size d, producing an array of shape [d, c], whose d-sized axis is
    layed out along the placement axis of the mesh, with layout of c-sized axis
    inherited from the operation of f. e.g., if f is jit-compiled with no
    sharding specifications, the layout of c will be determined by JAX's
    sharding propagation.

    In the case that the user has a mesh which does _not_ have placement as an
    axis name, the resulting array of shape [d, c], where the axis of size d is
    replicated across devices, and the axis of size c may be similarly sharded
    as above (depending on annotations internal to f, and the way the function
    constructed here is jit-compiled).

    The implementation here is intended to satisfy the sketch above (and its
    generalizations, including to pytrees, etc).

    Args:
      fn: Function to be mapped.
      arg: PyTree of arrays for which to map leading axis. If `placement` is not
        `'server'`, each array in the structure is assumed to have a leading
        axis of the same size, the number of elements at `placement`.
      placement: String representing the placement of input and output arrays to
        this map.

    Returns:
      The result of mapping `fn` over the leading axis (if `'placement'` is
      not `'server'`), satisfying the sharding requirements specified above.
    """
    # For server-placed values, there is no xmap/vmap dimension to speak of.
    # Attempting to use these on a non-existent dimension can lead to some
    # corner cases, when all we want is to evaluate the original fn.
    if placement == SERVER_PLACEMENT:
      mapped_fn = fn
    else:
      # `spmd_axis_name`` causes the `vmap`` to respect the sharding along this
      # axis name.
      if _placement_axis_in_mesh(placement):
        mapped_fn = jax.vmap(
            fn, axis_name=placement, in_axes=0, spmd_axis_name=placement
        )
      else:
        # Users should be free to use whatever mesh their model needs without
        # _necessarily_ registering a mesh-dimension for every placement with
        # which they are programming.
        mapped_fn = jax.vmap(
            fn,
            axis_name=placement,
            in_axes=0,
        )

    return call_jaxpr(
        mapped_fn,
        arg,
    )



# coding=utf-8
# Copyright 2024 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import parameterized
import jax
from jax import numpy as jnp
import tensorflow as tf

from . import impls
from . import primitives


def _jaxpr_has_primitive(jaxpr, prim_name):
  """A reimplementation of the fun of the same name in jax._src_dispatch."""
  for eqn in jaxpr.eqns:
    if prim_name in eqn.primitive.name:
      return True
    for subjaxpr in jax.core.subjaxprs(jaxpr):
      if _jaxpr_has_primitive(subjaxpr, prim_name):
        return True
  return False


class PrimitivesActingOnArraysTest(tf.test.TestCase, parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._n_clients = 100
    self._impl_defs = impls.PlacedComputations(
        {'clients': self._n_clients},
    )
    self._primdefs, _ = primitives.register_primitives(
        {'clients': self._n_clients},
    )

  def test_broadcast_clients_evaluation(self):
    fn = self._primdefs['broadcast_clients']
    # Check that this function is callable.
    self.assertAllClose(fn(jnp.array(1.0)), jnp.ones(shape=[self._n_clients]))
    # Check that it's jittable.
    self.assertAllClose(
        jax.jit(fn)(jnp.array(1.0)), jnp.ones(shape=[self._n_clients])
    )
    # Check that its forward-diffable.
    self.assertAllClose(
        jax.jacfwd(fn)(jnp.array(1.0)), jnp.ones(shape=[self._n_clients])
    )
    # Also that it's reverse-diffable.
    self.assertAllClose(
        jax.jacrev(fn)(jnp.array(1.0)), jnp.ones(shape=[self._n_clients])
    )

  def test_broadcast_clients_closure_under_fad(self):
    fn = self._primdefs['broadcast_clients']
    # Check that the forward and reverse-mode derivatives generate the expected
    # primitives.
    fwd_mode_jaxpr = jax.make_jaxpr(jax.jacfwd(fn))(jnp.array(1.0))
    self.assertTrue(_jaxpr_has_primitive(fwd_mode_jaxpr, 'broadcast_clients'))
    rev_mode_jaxpr = jax.make_jaxpr(jax.jacrev(fn))(jnp.array(1.0))
    self.assertTrue(_jaxpr_has_primitive(rev_mode_jaxpr, 'sum_from_clients'))

  def test_sum_from_clients_evaluation(self):
    fn = self._primdefs['sum_from_clients']
    clients_ones = jnp.ones(shape=[self._n_clients, 1])
    # Check that this function is callable.
    self.assertAllClose(fn(clients_ones), jnp.array([1.0 * self._n_clients]))
    # Check that it's jittable.
    self.assertAllClose(
        jax.jit(fn)(clients_ones), jnp.array([1.0 * self._n_clients])
    )
    # Check that its forward-diffable.
    self.assertAllClose(
        jax.jacfwd(fn)(clients_ones), jnp.ones(shape=[1, 100, 1])
    )
    # Check that its reverse-diffable.
    self.assertAllClose(
        jax.jacrev(fn)(clients_ones), jnp.ones(shape=[1, self._n_clients, 1])
    )

  def test_broadcast_and_sum_from_clients_eval(self):
    fn = self._primdefs['sum_from_clients']

    def _broadcast_then_sum(x):
      broadcasted_x = self._primdefs['broadcast_clients'](x)
      return fn(broadcasted_x)

    # This thing corresponds to fwd-mode AD in our paper.
    self.assertAllClose(
        jax.jacfwd(_broadcast_then_sum)(jnp.array([1.0])),
        jnp.array([[1.0 * self._n_clients]]),
    )

    # And here's reverse-ad.
    self.assertAllClose(
        jax.jacrev(_broadcast_then_sum)(jnp.array([1.0])),
        jnp.array([[1.0 * self._n_clients]]),
    )

  def test_sum_from_clients_closure_under_fad(self):
    # Check that the forward and reverse-mode derivatives generate the expected
    # primitives.
    fn = self._primdefs['sum_from_clients']
    clients_ones = jnp.ones(shape=[self._n_clients, 1])
    fwd_mode_jaxpr = jax.make_jaxpr(jax.jacfwd(fn))(clients_ones)
    self.assertTrue(_jaxpr_has_primitive(fwd_mode_jaxpr, 'sum_from_clients'))
    rev_mode_jaxpr = jax.make_jaxpr(jax.jacrev(fn))(clients_ones)
    self.assertTrue(_jaxpr_has_primitive(rev_mode_jaxpr, 'broadcast_clients'))

  def test_mean_from_clients_eval(self):
    fn = self._primdefs['mean_from_clients']
    clients_ones = jnp.ones(shape=[self._n_clients, 1])
    # Check that this function is callable.
    self.assertAllClose(fn(clients_ones), jnp.array([1.0]))
    # Check that it's jittable.
    self.assertAllClose(jax.jit(fn)(clients_ones), jnp.array([1.0]))
    # Check that its forward-diffable.
    self.assertAllClose(
        jax.jacfwd(fn)(clients_ones),
        1 / self._n_clients * jnp.ones(shape=[1, self._n_clients, 1]),
    )

  def test_broadcast_then_mean_from_clients_eval(self):
    fn = self._primdefs['mean_from_clients']

    def _broadcast_then_sum(x):
      broadcasted_x = self._primdefs['broadcast_clients'](x)
      return fn(broadcasted_x)

    # Again, let's do the forward-mode, reverse-mode checks.
    self.assertAllClose(
        jax.jacfwd(_broadcast_then_sum)(jnp.array([1.0])),
        jnp.array([[1.0]]),
    )
    self.assertAllClose(
        jax.jacrev(_broadcast_then_sum)(jnp.array([1.0])),
        jnp.array([[1.0]]),
    )

  def test_mean_from_clients_closure_under_fad(self):
    # Check that the forward and reverse-mode derivatives generate the expected
    # primitives.
    fn = self._primdefs['mean_from_clients']
    clients_ones = jnp.ones(shape=[self._n_clients, 1])
    fwd_mode_jaxpr = jax.make_jaxpr(jax.jacfwd(fn))(clients_ones)
    self.assertTrue(_jaxpr_has_primitive(fwd_mode_jaxpr, 'mean_from_clients'))
    rev_mode_jaxpr = jax.make_jaxpr(jax.jacrev(fn))(clients_ones)
    self.assertTrue(_jaxpr_has_primitive(rev_mode_jaxpr, 'broadcast_clients'))

  @parameterized.named_parameters(
      (
          'broadcast',
          'broadcast_clients',
          lambda _: jnp.array(1.0),
          lambda n: jnp.ones(shape=[n]),
      ),
      (
          'sum',
          'sum_from_clients',
          lambda n: jnp.ones(shape=[n, 1]),
          lambda n: jnp.ones(shape=[1, n, 1]),
      ),
      (
          'mean',
          'mean_from_clients',
          lambda n: jnp.ones(shape=[n, 1]),
          lambda n: 1 / n * jnp.ones(shape=[1, n, 1]),
      ),
  )
  def test_broadcast_clients_reverse_ad_with_symbolic_zero_and_jit(
      self, prim_name, arg_fn, result_fn
  ):
    fn = self._primdefs[prim_name]

    @jax.jit
    def duplicate_prim_result(x):
      return fn(x), fn(x)

    @jax.jit
    def ignore_prim_result(x):
      # Ignoring one result from this tuple-returning function triggers
      # reverse evaluation with a symbolic zero cotangent argument.
      y, _ = duplicate_prim_result(x)
      return y

    jac = jax.jacrev(ignore_prim_result)
    self.assertAllClose(
        jac(arg_fn(self._n_clients)), result_fn(self._n_clients)
    )


if __name__ == '__main__':
  tf.test.main()
